#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <time.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <getopt.h>
#include <assert.h>
#include <sys/wait.h>
#include <SDL.h>
#include <SDL_opengl.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "version.h"
#include "panic.h"
#include "list.h"
#include "render.h"
#include "vector.h"
#include "matrix.h"
#include "mesh.h"
#include "bsptree.h"
#include "model.h"
#include "vrml.h"
#include "list.h"
#include "cylinder.h"
#include "pick.h"
#include "move.h"
#include "game.h"
#include "engine.h"
#include "font.h"
#include "ui.h"

#define MAX(a, b) ((a) > (b) ? (a) : (b))
#define VBUTTON(x) ((x) == SDL_BUTTON_LEFT ? BTN_LEFT : (x) == SDL_BUTTON_RIGHT ? (BTN_RIGHT) : 0)

enum {
	WIDTH = 512,			/* window start width */
	HEIGHT = 512,			/* window start height */
	DEFAULT_MAX_DEPTH = 4,		/* default AI search depth */
	CHILD_WAIT_TIMEOUT = 1,		/* in seconds */
};

static struct worker_request {
	struct board_state state;
	int side;
	int max_depth;
} worker_request;

static struct {
	SDL_Thread *thread;
	union move next_move;
	int request_ready;
	SDL_mutex *request_mutex;
	SDL_cond *request_cond;
} worker_thread;

static int sdl_flags;

unsigned long
msecs(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);

	return
	  (unsigned long)tv.tv_sec*1000UL +
	  (unsigned long)tv.tv_usec/1000UL;
}

void
swap_buffers(void)
{
	SDL_GL_SwapBuffers();
}

static void
sys_sigblock(int sig)
{
	sigset_t mask;

	sigemptyset(&mask);
	sigaddset(&mask, sig);
	sigprocmask(SIG_BLOCK, &mask, (sigset_t *)NULL);
}


static void
sys_sigrelease(int sig)
{
	sigset_t mask;

	sigemptyset(&mask);
	sigaddset(&mask, sig);
	sigprocmask(SIG_UNBLOCK, &mask, (sigset_t *)NULL);
}

static
void (*sys_sigset(int sig, void(*func)(int)))(int)
{
	struct sigaction act, oact;

	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	act.sa_handler = func;
	sigaction(sig, &act, &oact);

	return oact.sa_handler;
}

static int 
worker_loop(void *dummy)
{
	int done = 0;

	while (!done) {
		struct worker_request req;
		SDL_Event event;

		SDL_mutexP(worker_thread.request_mutex);

		while (!worker_thread.request_ready)
			SDL_CondWait(worker_thread.request_cond,
			  worker_thread.request_mutex);

		done = worker_thread.request_ready == -1;

		worker_thread.request_ready = 0;

		if (!done)
			req = worker_request;

		SDL_mutexV(worker_thread.request_mutex);

		if (!done) {
			get_best_move(&worker_thread.next_move, &req.state,
			  req.side, req.max_depth);

			event.type = SDL_USEREVENT;
			SDL_PushEvent(&event);
		}
	}

	return 0;
}

static void
init_worker(void)
{
	worker_thread.request_cond = SDL_CreateCond();
	worker_thread.request_mutex = SDL_CreateMutex();
	worker_thread.request_ready = 0;

	worker_thread.thread = SDL_CreateThread(worker_loop, NULL);
}

static void
kill_worker(void)
{
	int dummy;

	SDL_mutexP(worker_thread.request_mutex);

	worker_thread.request_ready = -1;

	SDL_CondSignal(worker_thread.request_cond);
	SDL_mutexV(worker_thread.request_mutex);

	SDL_WaitThread(worker_thread.thread, &dummy);
}

int
send_worker_request(void)
{
	SDL_mutexP(worker_thread.request_mutex);

	worker_request.state = ui_context.state;
	worker_request.side = ui_context.cur_side;
	worker_request.max_depth = ui_context.max_depth;

	worker_thread.request_ready = 1;

	SDL_CondSignal(worker_thread.request_cond);
	SDL_mutexV(worker_thread.request_mutex);

	return 0;
}

#if 0
static void
on_key_press(struct ui_context *ctx, XKeyEvent *xkey)
{
	int code;

	code = XLookupKeysym(xkey, 0);

	switch (code) {
		case XK_Control_L:
		case XK_Control_R:
			ctx->control_key_down = 1;
			break;
	}
}

static void
on_key_release(struct ui_context *ctx, XKeyEvent *xkey)
{
	int code;

	code = XLookupKeysym(xkey, 0);

	switch (code) {
		case XK_Control_L:
		case XK_Control_R:
			ctx->control_key_down = 0;
			break;
	}
}

static int
on_x_event(struct ui_context *ctx)
{
	XEvent event;
	int done;

	done = 0;

	while (XPending(the_display)) {
		XNextEvent(the_display, &event);

		switch (event.type) {
			case ClientMessage:
				if ((Atom)(event.xclient.data.l[0]) ==
				  close_atom)
					done = 1;
				break;

			case ButtonPress:
				on_button_press(ctx,
				  VBUTTON(event.xbutton.button),
				  event.xbutton.x, event.xbutton.y);
				break;

			case ButtonRelease:
				on_button_release(ctx,
				  VBUTTON(event.xbutton.button),
				  event.xbutton.x, event.xbutton.y);
				break;

			case KeyPress:
				on_key_press(ctx, &event.xkey);
				break;

			case KeyRelease:
				on_key_release(ctx, &event.xkey);
				break;

			case MotionNotify:
				while (XCheckTypedWindowEvent(the_display,
				  the_window, MotionNotify, &event))
					;

				on_motion_notify(ctx,
				  VBUTTON(event.xbutton.button),
				  event.xbutton.x, event.xbutton.y);
				break;

			case Expose:
				ui_redraw(ctx);
				break;

			case ConfigureNotify:
				ui_reshape(ctx, event.xconfigure.width,
				  event.xconfigure.height);
				break;
		}
	}

	return done;
}

static void
on_x_worker_reply(struct ui_context *ctx)
{
	union move move = {0};

	if (read_exact(the_worker.read_from, &move, sizeof move) !=
	  sizeof move)
		panic("invalid data from worker");

	on_worker_reply(ctx, &move);
}
#endif

void
event_loop(void)
{
	int done;
	SDL_Event event;

	ui_reshape(WIDTH, HEIGHT);

	done = 0;

	while (!done) {
		ui_on_idle();

		while (SDL_PollEvent(&event)) {
			switch (event.type) {
				case SDL_MOUSEBUTTONDOWN:
					ui_on_button_press(
					  VBUTTON(event.button.button),
					  event.button.x, event.button.y);
					break;

				case SDL_MOUSEBUTTONUP:
					ui_on_button_release(
					  VBUTTON(event.button.button),
					  event.button.x, event.button.y);
					break;

				case SDL_MOUSEMOTION:
					ui_on_motion_notify(
					  VBUTTON(event.button.button),
					  event.button.x, event.button.y);
					break;

				case SDL_KEYDOWN:
					if (event.key.keysym.sym == SDLK_ESCAPE)
						done = 1;
					break;

				case SDL_VIDEORESIZE:
					ui_reshape(event.resize.w,
					  event.resize.h);
					SDL_SetVideoMode(event.resize.w,
					  event.resize.h, 0, sdl_flags);
					break;

				case SDL_USEREVENT:
					/* answer from worker */
					ui_on_worker_reply(
					  &worker_thread.next_move);
					break;

				case SDL_QUIT:
					done = 1;
					break;

				default:
					break;
			}
		}

		ui_redraw();

		swap_buffers();
	}
}

#if 0
static Window
make_window(Display *display, int width, int height)
{
	int attrib[] = { GLX_RGBA,
		    GLX_RED_SIZE, 1,
		    GLX_GREEN_SIZE, 1,
		    GLX_BLUE_SIZE, 1,
		    GLX_DOUBLEBUFFER,
		    GLX_DEPTH_SIZE, 1,
		    None };
	int screen;
	XSetWindowAttributes attr;
	unsigned long mask;
	Window root;
	Window window;
	GLXContext ctx;
	XVisualInfo *visinfo;

	screen = DefaultScreen(display);
	root = RootWindow(display, screen);

	if ((visinfo = glXChooseVisual(display, screen, attrib)) == NULL)
	   panic("couldn't get RGB, double-buffered visual");

	/* window attributes */
	attr.background_pixel = 0;
	attr.border_pixel = 0;
	attr.colormap = XCreateColormap(display, root, visinfo->visual,
	  AllocNone);
	attr.event_mask = StructureNotifyMask|ExposureMask|ButtonPressMask|
	  ButtonReleaseMask|KeyPressMask|KeyReleaseMask|PointerMotionMask;
	mask = CWBackPixel|CWBorderPixel|CWColormap|CWEventMask;

	window = XCreateWindow(display, root, 0, 0, width, height, 0,
	  visinfo->depth, InputOutput, visinfo->visual, mask, &attr);

	if ((ctx = glXCreateContext(display, visinfo, NULL, True)) == NULL)
	   panic("glXCreateContext failed");

	glXMakeCurrent(display, window, ctx);

	return window;
}
#endif

static void
initialize_sdl(void)
{
	if (SDL_Init(SDL_INIT_VIDEO) < 0)
		panic("SDL_Init: %s", SDL_GetError());

	if (SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1) < 0)
		panic("SDL_GL_SetAttribute: %s", SDL_GetError());

	sdl_flags = SDL_OPENGL|SDL_GL_DOUBLEBUFFER|SDL_RESIZABLE;

	if (SDL_SetVideoMode(WIDTH, HEIGHT, 0, sdl_flags) == NULL)
		panic("SDL_SetVideoMode: %s", SDL_GetError());

	SDL_WM_SetCaption(VERSION, NULL);
}

static void
usage(void)
{
	fprintf(stderr, VERSION"\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "  Usage: vulcan [options]\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "Options are:\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "  -c   computer vs. computer\n");
	fprintf(stderr, "  -u   human vs. human\n");
	fprintf(stderr, "  -b   play black\n");
	fprintf(stderr, "  -d   set maximum AI search depth\n");

	exit(1);
}

int
main(int argc, char *argv[])
{
	int c;
	enum player_type white_player, black_player;
	int max_depth;
	char *p;

	white_player = HUMAN_PLAYER;
	black_player = COMPUTER_PLAYER;
	max_depth = DEFAULT_MAX_DEPTH;

	while ((c = getopt(argc, argv, "cubd:h")) != EOF) {
		switch (c) {
			case 'c':
				white_player = black_player = COMPUTER_PLAYER;
				break;

			case 'u':
				white_player = black_player = HUMAN_PLAYER;
				break;

			case 'b':
				white_player = COMPUTER_PLAYER;
				black_player = HUMAN_PLAYER;
				break;

			case 'd':
				max_depth = strtol(optarg, &p, 10);
				if (p == optarg || max_depth < 1)
					usage();
				break;

			case 'h':
			default:
				usage();
		}
	}

	ui_initialize(white_player, black_player, max_depth);
	/* init_signals(); */
	init_worker();

	initialize_sdl();
	gl_initialize();

	event_loop();

	kill_worker();

	SDL_Quit();

	return 0;
}
