#include <stdlib.h>
#include "map.h"

struct map_entry {
	void *key;
	void *value;
	struct map_entry *next;
};

struct map {
	int nbuckets;
	struct map_entry **buckets;
	unsigned (*key_hash)(const void *key);
	int (*key_equals)(const void *key1, const void *key2);
};

struct map *
map_make(int nbuckets, unsigned (*key_hash)(const void *key),
  int (*key_equals)(const void *key1, const void *key2))
{
	struct map *map = malloc(sizeof *map);

	map->nbuckets = nbuckets;
	map->buckets = calloc(map->nbuckets, sizeof *map->buckets);
	map->key_hash = key_hash;
	map->key_equals = key_equals;

	return map;
}

static struct map_entry *
map_entry_make(void *key, void *value, struct map_entry *next)
{
	struct map_entry *p = malloc(sizeof *p);

	p->key = key;
	p->value = value;
	p->next = next;

	return p;
}

static int
bucket_idx(struct map *map, const void *key)
{
	return map->key_hash(key)%map->nbuckets;
}

void
map_put(struct map *map, void *key, void *value)
{
	int idx = bucket_idx(map, key);

	map->buckets[idx] = map_entry_make(key, value, map->buckets[idx]);
}

void *
map_get(struct map *map, const void *key)
{
	struct map_entry *p;

	for (p = map->buckets[bucket_idx(map, key)]; p; p = p->next) {
		if (key == p->key || map->key_equals(key, p->key))
			return p->value;
	}

	return NULL;
}

void
map_free(struct map *map, void (*free_key)(void *), void (*free_data)(void *))
{
	int i;

	for (i = 0; i < map->nbuckets; i++) {
		struct map_entry *p, *n;

		for (p = map->buckets[i]; p; p = n) {
			n = p->next;

			free_key(p->key);
			free_data(p->value);
			free(p);
		}
	}
}
